// Pingus - A free Lemmings clone
// Copyright (C) 1999 Ingo Ruhnke <grumbel@gmx.de>
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

#include <iostream>
#include <cstdio>
#include <string>
#include <fstream>

#include "PLF.hh"
#include "globals.hh"
map<string, MakeAction> PLF::action_map;

PLF::PLF()
{
  x_pos = y_pos = 0;
  pingus = 50;
}

// Takes the name of the PinguLevelFile as argument and reads its
// data, to make it easy accessible for the other objects.
PLF::PLF(string filename)
{
  music = "march.it";

  string token;
  string dir;

  maptype = UNDEF;
  x_pos = y_pos = 0;

  ifstream in(filename.c_str());

  if (!in) {
    perror(filename.c_str());
    exit(EXIT_FAILURE);
  }

  // Get directory
  dir = filename;
  for(int i=dir.size(); i>=0; --i) {
    if (dir[i] == '/') {
      dir[i+1] = '\0';
      break;
    }
  }

  bg  = new char[1024];
  fg  = new char[1024];
  col = new char[1024];
  string tmp;

  fg[0] = '\0';
  strcpy(col, dir.c_str());

  hotspot_t spot;
  exit_data exit_p;
  entrance_data entrance_p;
  trap_data trap;

  while (!in.eof()) {
    in >> token;

    if (in.eof()) {
      if (verbose) cout << "PLF: found EOF" << endl;
      break;
    }

    if (verbose > 4)
      cout << " " << token;

    switch (get_token(token.c_str())) {

    case BG:
      // Reads the background an its animation speed (-1 for no animation)
      in >> bg;
      in >> bg_speed;
      break;

    case FG:
      // Reads the tile map name and corrects its path
      in >> tmp;
      
      if (maptype != BMP)
	strcpy(fg, dir.c_str());
      
      strcat(fg, tmp.c_str());
      break;

    case COL:
      // Reads the colisionmap name and corrects the path
      in >> tmp;
      strcat(col, tmp.c_str());
      if (verbose)
	cout << "PLF:" << col << endl; 
      break;

    case EMPTY:
      if (verbose)
	cout << "PLF: Empty token!" << endl;
      break;

    case COMMENT:
      // Jumps to the end of the current line
      in.ignore(255, '\n');
      break;

    case LONG_COMMENT:
      while(true) {
	in >> tmp;
	if (tmp == "*/")
	  break;
      }
      break;

    case START:
      // Reads the start position of the camera in the level
      in >> x_pos;
      in >> y_pos;
      break;

    case HOTSPOT_BG:
      in >> spot.name;
      in >> spot.x_pos;
      in >> spot.y_pos;
      hotspot_bg.push_back(spot);
      break;

    case HOTSPOT_FG:
      in >> spot.name;
      in >> spot.x_pos;
      in >> spot.y_pos;
      hotspot_fg.push_back(spot);
      break;

    case EXIT:
      in >> exit_p.name;
      in >> exit_p.x_pos;
      in >> exit_p.y_pos;
      exit_v.push_back(exit_p);
      break;

    case ENTRANCE:
      {
	string t;
	in >> entrance_p.name;
	in >> entrance_p.x_pos;
	in >> entrance_p.y_pos;

	in >> t;

	if (t == "left")
	  entrance_p.direction.left();
	else 
	  entrance_p.direction.right();

	in >> entrance_p.release_rate;
	entrance_v.push_back(entrance_p);
      }
      break;

    case MAPTYPE:
      in >> token;
      if (token == "BMP") {
	maptype = BMP;
      } else if (token == "TILE") {
	maptype = TILE;
      } else if (token == "SPOT") {
	maptype = SPOT;
      } else if (token == "RANDOM") {
	maptype = RANDOM;
      } else {
	cout << "Syntax Error in PLF file! Wrong maptyp!" << endl;
	exit(EXIT_FAILURE);
      }
      break;
    case TRAP:
      in >> trap.name;
      in >> trap.x_pos;
      in >> trap.y_pos;
      in >> trap.data;
      traps.push_back(trap);
      break;

    case PINGUS:
      in >> pingus;
      break;

    case MUSIC:
      in >> music;
      break;

    case DESCRIPTION:
      {
	string temp;
	while(temp != "DESCRIPTION_END" && !in.eof()) {
	  in >> temp;
	  description += temp;
	}
      }
      break;

    case BUTTONS:
      {
	string   name;
	int number_of;
	button_data button;

	in >> name;
	while (name != "BUTTONS_END") {
	  if (in.eof()) {
	    cout << "Buttonname: `" << name << "'" << endl;
	    cout << "Unexpected EOF" << endl;
	    break;
	  }
	  button.name = name;
	  button.make_action = action_map[button.name];
	  in >> number_of;
	  button.number_of = number_of;
	  
	  buttons.push_back(button);
	  in >> name;
	  if (verbose > 4) cout << " " << name;
	}
	if (verbose > 4) cout << "Exited Button loop" << endl;
      }
      break;
    case UNKNOWN:
      cout << "Syntax Error in PLF file!" << endl;
      exit(EXIT_FAILURE);
      break;
    }
  }
  
  if (maptype == UNDEF) {
    cout << "No maptype set in PLF-file" << endl;
    exit(EXIT_FAILURE);
  }
}

plf_token
PLF::get_token(string token)
{
  // toupper(token);

  if (token == "BG:") {
    return BG;
  } else if (token == "FG:") {
    return FG;
  } else if (token == "COL:") {
    return COL;
  } else if (token == "MAPTYPE:") {
    return MAPTYPE;
  } else if (token == "") {
    return EMPTY;
  } else if (token == "//") {
    return COMMENT;
  } else if (token == "/*") {
    return LONG_COMMENT;
  } else if (token == "START:") {
    return START;
  } else if (token == "HOTSPOT_FG:") {
    return HOTSPOT_FG;
  } else if (token == "HOTSPOT_BG:") {
    return HOTSPOT_BG;
  } else if (token == "EXIT:") {
    return EXIT;
  } else if (token == "ENTRANCE:") {
    return ENTRANCE;
  } else if (token == "TRAP:") {
    return TRAP;
  } else if (token == "PINGUS:") {
    return PINGUS;
  } else if (token == "BUTTONS_BEGIN") {
    return BUTTONS;
  } else if (token == "DESCRIPTION_BEGIN") {
    return DESCRIPTION;
  } else if (token == "MUSIC:") {
    return MUSIC;
  }

  cout << "Current Token: " << token << endl;
  return UNKNOWN;
}

// A list of functions, which makes the private variables readable for 
// the outerworld.

const char*
PLF::get_bg()
{
  return bg;
}

const char*
PLF::get_fg()
{
  return fg;
}

const char*
PLF::get_pngfile()
{
  return "BX";
  //  return (filename + ".png").c_str();  
}

MapType
PLF::map_type()
{
  return maptype;
}

char const*
PLF::get_mapfile()
{
  return col;
}

int
PLF::get_startx()
{
  return -x_pos;
}

int
PLF::get_starty()
{
  return y_pos;
}

vector<hotspot_t>
PLF::get_hotspot_fg()
{
  return hotspot_fg;
}

vector<hotspot_t>
PLF::get_hotspot_bg()
{
  return hotspot_bg;
}

vector<exit_data>
PLF::get_exit()
{
  return exit_v;
}

vector<entrance_data>
PLF::get_entrance()
{
  return entrance_v;
}

vector<trap_data> 
PLF::get_traps()
{
  return traps;
}

int
PLF::get_pingus()
{
  return pingus;
}

vector<button_data>
PLF::get_buttons()
{
  return buttons;
}

// A function which regiersters a new button, so that it can be used
// in a level file between BUTTON_BEGIN and BUTTON_END
void
PLF::register_new_button(string name, MakeAction func)
{
  action_map[name] = func;
}

string
PLF::get_music()
{
  return music;
}

/* EOF */
