// Pingus - A free Lemmings clone
// Copyright (C) 1999 Ingo Ruhnke <grumbel@gmx.de>
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

#include <cassert>
#include <ClanLib/clanlib.h>
#include <ClanLib/clanlayer2.h>

#include "globals.hh"
#include "Pingu.hh"
#include "actions/pingu_actions.hh"
#include "actions/smashed.hh"

// Create the static objects of the class
int Pingu::id_count;

ColMap*             Pingu::colmap;
PinguMap*           Pingu::map;

CL_Surface*         Pingu::walker;
CL_Surface*         Pingu::faller;
CL_Surface*         Pingu::tumble;
CL_Font*            Pingu::font;

// Init a pingu at the given position while falling
Pingu::Pingu(int x, int y)
{
  id = id_count++;

  action = sec_action = 0;
  action_time = -1;
  data.x_pos = x;
  data.y_pos = y;

  walker_c.set_size(walker->get_no_sprs() / 2);
  walker_c.set_speed(75);
  
  faller_c.set_size(faller->get_no_sprs());
  tumble_c.set_size(tumble->get_no_sprs());

  data.falling = 4;
  data.direction.left();
}

// Returns the x position of the pingu
int
Pingu::get_x()
{
  return data.x_pos;
}

// Returns the y position of the pingu
int
Pingu::get_y()
{
  return data.y_pos;
}

int
Pingu::get_id()
{
  return id;
}

//Returns environment
PinguEnvironment
Pingu::get_environment()
{
  return data.environment;
}

// Set the position of the pingu
void 
Pingu::set_pos(int x, int y)
{
  data.x_pos = x;
  data.y_pos = y;
}

// Static function which set the colmap for all pingus
void
Pingu::set_colmap(ColMap* a)
{
  colmap = a;
}

// Static function which set the map for all pingus
void
Pingu::set_map(PinguMap* a)
{
  map = a;
}

// Static function which load the default surfaces for all pingus
void 
Pingu::load_data()
{
  if (verbose > 2)
    cout << __FILE__ << ":" << __LINE__ << ": Loading data" << endl;
  
  font = CL_Res_Font::load("Fonts/numbers",*resource());  
  walker = CL_Res_Surface::load("Pingus/walker", *resource());
  faller = CL_Res_Surface::load("Pingus/faller", *resource());
  tumble = CL_Res_Surface::load("Pingus/tumble", *resource());

  if (verbose > 2)
    cout << __FILE__ << ":" << __LINE__ << ": Loading data successfull" << endl;
}

// Set the action of the pingu (bridger, blocker, bomber, etc.)
// The `t' parameter set the amount of time which have to be passed
// before the action is set. A value of `-1' activates the action
// directly.
int
Pingu::set_action(PinguAction* act, int t)
{
  assert(act);

  {
    char str[256];
    sprintf(str, "Pingu[%d]::set_action(%s)", id, act->name());
    event.send(str);
  }
  
  // Set the colmap and the map
  act->set_colmap(colmap);
  act->set_map(map);

  if (t == -2)
    t = act->activation_time();
  
  if (t == -1) {
    if (!act)
      return 0;
    
    if (!(act->get_environment() & data.environment)){  // if environment is bad;
      delete act;
      return 0; 
    }
    
    if (action)
      delete action;
    
    action = act;
    action->set_pingu_data(data);
    action->load_data();
    action_time = -1;
    return 1;
  } else {
    action_time = t;
    
    if (sec_action)
      delete sec_action;
    
    sec_action = act;
  }
  return true;
}

PinguStatus
Pingu::get_status(void)
{
  return data.status;
}

// Returns true if the given koordinates are above the pingu
bool
Pingu::is_over(int x, int y)
{
  if (data.status == (dead || exited) || data.status == not_catchable)
    return false;

  if (data.x_pos - 16 < x && data.x_pos + 16 > x &&
      data.y_pos - 32 < y && data.y_pos > y) {
    return true;
  } else {
    return false;
  }
}

// Let the pingu do his job (i.e. walk his way)
void
Pingu::let_move(void)
{
  assert(colmap);
  //  action->set_colmap(colmap); //BUG: the colmap need only to be set once
  //  action->set_map(map);
  
  if (action_time > -1) 
    --action_time;

  if (action_time == 0 && sec_action) {
    // BUG: Memory leak
    if (action) delete action;
    action = sec_action;
    action->set_pingu_data(data);
    action->load_data();
  }
  
  if (action && action->is_finished) {
    delete action;
    action = 0;
  }
    
  if (rel_getpixel(0, -1) == nothing) {
    data.status = dead;
    return;
  }

  if (action) { // if it has a action, do it
    action->let_move();
  } else {      // if not, do a walk
    if (rel_getpixel(0, -1) == air) { // if pingu is not on ground
      ++(data.y_pos);
      if (rel_getpixel(0, -1) == air) 
	++(data.y_pos);
      ++(data.falling);
      data.environment = sky ; 
    } else { // if pingu is on ground
      if (data.falling > 75)
        set_action(new Smashed);
      data.falling = 0;
      data.environment = land; 
      if (rel_getpixel(1, 0) == air) { // if infront is free
	data.x_pos += data.direction;    
      } else { // if infront is a pixel 
	// Pingu is walking up/down mountain 
	if (rel_getpixel(1,1) == air) {
	  data.x_pos += data.direction;
	  --data.y_pos;
	} else if (rel_getpixel(1,2) == air) {
	  data.x_pos += data.direction;
	  data.y_pos -= 2;
	} else {
	  data.direction.change();
	}
      }
    }
  }
}

// Draws the pingu on the screen with the given offset
void
Pingu::draw_offset(int x, int y, double s = 1.0)
{
  char str[64];
  y += 2;
  
  if (action) {
    action->draw_offset(x, y,s);
  } else {
    if (data.falling > 3) {
      CL_Surface* surf;
      
      if (data.falling > 75)
	surf = tumble;
      else
	surf = faller;
	
      if (s == 1.0) {
	surf->put_screen(data.x_pos + x - 16, data.y_pos + y - 32,
			 ++tumble_c);
      } else {
	surf->put_screen((data.x_pos + x - 16) * s , (data.y_pos + y - 32) * s,
			 (int)(tumble->get_width()*s), (int)(tumble->get_height()*s),
			 ++tumble_c);
      }
    } else {
      if (s == 1.0) {
	walker->put_screen(data.x_pos + x - 16, data.y_pos + y - 32, 
			   ++walker_c
			   + ((data.direction.is_left() ? 0 :
			       walker_c.size())));
      } else {
	walker->put_screen((data.x_pos + x - 16) * s, (data.y_pos + y - 32) * s, 
			   (int)(tumble->get_width()*s), (int)(tumble->get_height()*s),
			   ++walker_c
			   + ((data.direction.is_left() ? 0 :
			       walker_c.size())));
      }
    }
  }

  if (action_time != -1) {
    sprintf(str, "%d", action_time);
    if (s == 1.0) {
      font->print_center(data.x_pos + x, data.y_pos - 45 + y, str);
    } else if (s > 1.0) {
      // Bug?!: The font isn't zoomed
      font->print_center((int)((data.x_pos + x) * s), (int)((data.y_pos - 45 + y) * s), str);
    } else {
      // Don't draw the font if the zoom is smaller than default (1.0)
    }
  }
}

// Returns the `color' of the colmap in the walking direction 
int
Pingu::rel_getpixel(int x, int y)
{
  assert(colmap);
  return colmap->getpixel(data.x_pos + (x * data.direction), (data.y_pos) - y);
}

// Let the pingu catch another pingu, so that an action can be aplied
// (i.e. let a blocker change the direction f another pingu)
void
Pingu::catch_pingu(Pingu* pingu)
{
  action->catch_pingu(&(pingu->data));
}

// Returns true if the pingu needs to catch another pingu
bool
Pingu::need_catch()
{
  if (action)
    return action->need_catch();
  else 
    return false;
}

void
Pingu::set_direction(Direction d)
{
  data.direction = d;
}

bool
Pingu::is_alive(void)
{
  return (data.status != dead);
}

PinguAction*
Pingu::get_action()
{
  return action;
}

/* EOF */
